""" You can kind of see this as the scope of `metadate` when you 'import metadate'
The following functions become available:
metadate.__project__
metadate.__version__
metadate.print_version
metadate.parse_date
"""
from datetime import datetime as _datetime

now = _datetime.now
date = _datetime.date
time = _datetime.time


from metadate.parse_date import parse_date
from metadate.utils import Units
from metadate.classes import MetaPeriod
from dateutil.relativedelta import relativedelta as rd

import sys

__project__ = "metadate"
__version__ = "0.5.74"


def is_mp(x):
    return isinstance(x, MetaPeriod)


def print_version():
    """ Convenient function for printing the version of the package """
    sv = sys.version_info
    py_version = "{}.{}.{}".format(sv.major, sv.minor, sv.micro)
    version_parts = __version__.split(".")
    s = "{} version: [{}], Python {}".format(__project__, __version__, py_version)
    s += "\nMajor version: {}  (breaking changes)".format(version_parts[0])
    s += "\nMinor version: {}  (extra feature)".format(version_parts[1])
    s += "\nMicro version: {} (commit count)".format(version_parts[2])
    return s



TIMEZONES = [
    {
        "offset": 10.5,
        "name": "Australian Central Daylight Time",
        "code": "ACDT"
    },
    {
        "offset": 9.5,
        "name": "Australian Central Standard Time",
        "code": "ACST"
    },
    {
        "offset": -3.0,
        "name": "Atlantic Daylight Time",
        "code": "UNUSED_ADT"
    },
    {
        "offset": -3.0,
        "name": "Atlantic Daylight Time",
        "code": "ADT"
    },
    {
        "offset": 11.0,
        "name": "Australian Eastern Daylight Time",
        "code": "AEDT"
    },
    {
        "offset": 10.0,
        "name": "Australian Eastern Standard Time",
        "code": "AEST"
    },
    {
        "offset": 4.5,
        "name": "Afghanistan Time",
        "code": "AFT"
    },
    {
        "offset": -8.0,
        "name": "Alaska Daylight Time",
        "code": "AKDT"
    },
    {
        "offset": -9.0,
        "name": "Alaska Standard Time",
        "code": "AKST"
    },
    {
        "offset": 6.0,
        "name": "Alma-Ata Time",
        "code": "ALMT"
    },
    {
        "offset": 5.0,
        "name": "Armenia Summer Time",
        "code": "UNUSED_AMST"
    },
    {
        "offset": -3.0,
        "name": "Amazon Summer Time",
        "code": "AMST"
    },
    {
        "offset": 4.0,
        "name": "Armenia Time",
        "code": "UNUSED_AMT"
    },
    {
        "offset": -4.0,
        "name": "Amazon Time",
        "code": "AMT"
    },
    {
        "offset": 12.0,
        "name": "Anadyr Summer Time",
        "code": "ANAST"
    },
    {
        "offset": 12.0,
        "name": "Anadyr Time",
        "code": "ANAT"
    },
    {
        "offset": 5.0,
        "name": "Aqtobe Time",
        "code": "AQTT"
    },
    {
        "offset": -3.0,
        "name": "Argentina Time",
        "code": "ART"
    },
    {
        "offset": 3.0,
        "name": "Arabia Standard Time",
        "code": "UNUSED_AST"
    },
    {
        "offset": -4.0,
        "name": "Atlantic Standard Time",
        "code": "AST"
    },
    {
        "offset": 9.0,
        "name": "Australian Western Daylight Time",
        "code": "AWDT"
    },
    {
        "offset": 8.0,
        "name": "Australian Western Standard Time",
        "code": "AWST"
    },
    {
        "offset": 0.0,
        "name": "Azores Summer Time",
        "code": "AZOST"
    },
    {
        "offset": -1.0,
        "name": "Azores Time",
        "code": "AZOT"
    },
    {
        "offset": 5.0,
        "name": "Azerbaijan Summer Time",
        "code": "AZST"
    },
    {
        "offset": 4.0,
        "name": "Azerbaijan Time",
        "code": "AZT"
    },
    {
        "offset": 8.0,
        "name": "Brunei Darussalam Time",
        "code": "BNT"
    },
    {
        "offset": -4.0,
        "name": "Bolivia Time",
        "code": "BOT"
    },
    {
        "offset": -2.0,
        "name": "Brasilia Summer Time",
        "code": "BRST"
    },
    {
        "offset": -3.0,
        "name": "Brasília time",
        "code": "BRT"
    },
    {
        "offset": 6.0,
        "name": "Bangladesh Standard Time",
        "code": "BST"
    },
    {
        "offset": 1.0,
        "name": "British Summer Time",
        "code": "UNUSED_BST"
    },
    {
        "offset": 6.0,
        "name": "Bhutan Time",
        "code": "BTT"
    },
    {
        "offset": 8.0,
        "name": "Casey Time",
        "code": "CAST"
    },
    {
        "offset": 2.0,
        "name": "Central Africa Time",
        "code": "CAT"
    },
    {
        "offset": 6.5,
        "name": "Cocos Islands Time",
        "code": "CCT"
    },
    {
        "offset": -4.0,
        "name": "Cuba Daylight Time",
        "code": "UNUSED_CDT"
    },
    {
        "offset": -5.0,
        "name": "Central Daylight Time",
        "code": "CDT"
    },
    {
        "offset": 2.0,
        "name": "Central European Summer Time",
        "code": "CEST"
    },
    {
        "offset": 1.0,
        "name": "Central European Time",
        "code": "CET"
    },
    {
        "offset": 13.75,
        "name": "Chatham Island Daylight Time",
        "code": "CHADT"
    },
    {
        "offset": 12.75,
        "name": "Chatham Island Standard Time",
        "code": "CHAST"
    },
    {
        "offset": -10.0,
        "name": "Cook Island Time",
        "code": "CKT"
    },
    {
        "offset": -3.0,
        "name": "Chile Summer Time",
        "code": "CLST"
    },
    {
        "offset": -4.0,
        "name": "Chile Standard Time",
        "code": "CLT"
    },
    {
        "offset": -5.0,
        "name": "Colombia Time",
        "code": "COT"
    },
    {
        "offset": 8.0,
        "name": "China Standard Time",
        "code": "UNUSED_CST"
    },
    {
        "offset": -6.0,
        "name": "Central Standard Time",
        "code": "CST"
    },
    {
        "offset": -5.0,
        "name": "Cuba Standard Time",
        "code": "UNUSED_CST"
    },
    {
        "offset": -1.0,
        "name": "Cape Verde Time",
        "code": "CVT"
    },
    {
        "offset": 7.0,
        "name": "Christmas Island Time",
        "code": "CXT"
    },
    {
        "offset": 10.0,
        "name": "Chamorro Standard Time",
        "code": "ChST"
    },
    {
        "offset": 7.0,
        "name": "Davis Time",
        "code": "DAVT"
    },
    {
        "offset": -5.0,
        "name": "Easter Island Summer Time",
        "code": "EASST"
    },
    {
        "offset": -6.0,
        "name": "Easter Island Standard Time",
        "code": "EAST"
    },
    {
        "offset": 3.0,
        "name": "Eastern Africa Time",
        "code": "UNUSED_EAT"
    },
    {
        "offset": 3.0,
        "name": "East Africa Time",
        "code": "EAT"
    },
    {
        "offset": -5.0,
        "name": "Ecuador Time",
        "code": "ECT"
    },
    {
        "offset": -4.0,
        "name": "Eastern Daylight Time",
        "code": "EDT"
    },
    {
        "offset": 3.0,
        "name": "Eastern European Summer Time",
        "code": "EEST"
    },
    {
        "offset": 2.0,
        "name": "Eastern European Time",
        "code": "EET"
    },
    {
        "offset": 0.0,
        "name": "Eastern Greenland Summer Time",
        "code": "EGST"
    },
    {
        "offset": -1.0,
        "name": "East Greenland Time",
        "code": "EGT"
    },
    {
        "offset": -5.0,
        "name": "Eastern Standard Time",
        "code": "EST"
    },
    {
        "offset": 13.0,
        "name": "Fiji Summer Time",
        "code": "FJST"
    },
    {
        "offset": 12.0,
        "name": "Fiji Time",
        "code": "FJT"
    },
    {
        "offset": -3.0,
        "name": "Falkland Islands Summer Time",
        "code": "FKST"
    },
    {
        "offset": -4.0,
        "name": "Falkland Island Time",
        "code": "FKT"
    },
    {
        "offset": -2.0,
        "name": "Fernando de Noronha Time",
        "code": "FNT"
    },
    {
        "offset": -6.0,
        "name": "Galapagos Time",
        "code": "GALT"
    },
    {
        "offset": -9.0,
        "name": "Gambier Time",
        "code": "GAMT"
    },
    {
        "offset": 4.0,
        "name": "Georgia Standard Time",
        "code": "GET"
    },
    {
        "offset": -3.0,
        "name": "French Guiana Time",
        "code": "GFT"
    },
    {
        "offset": 12.0,
        "name": "Gilbert Island Time",
        "code": "GILT"
    },
    {
        "offset": 0.0,
        "name": "Greenwich Mean Time",
        "code": "GMT"
    },
    {
        "offset": 4.0,
        "name": "Gulf Standard Time",
        "code": "GST"
    },
    {
        "offset": -4.0,
        "name": "Guyana Time",
        "code": "GYT"
    },
    {
        "offset": -3.0,
        "name": "Heure Avancée de l'Atlantique",
        "code": "HAA"
    },
    {
        "offset": -5.0,
        "name": "Heure Avancée du Centre",
        "code": "HAC"
    },
    {
        "offset": -9.0,
        "name": "Hawaii-Aleutian Daylight Time",
        "code": "HADT"
    },
    {
        "offset": -4.0,
        "name": "Heure Avancée de l'Est",
        "code": "HAE"
    },
    {
        "offset": -7.0,
        "name": "Heure Avancée du Pacifique",
        "code": "HAP"
    },
    {
        "offset": -6.0,
        "name": "Heure Avancée des Rocheuses",
        "code": "HAR"
    },
    {
        "offset": -10.0,
        "name": "Hawaii-Aleutian Standard Time",
        "code": "HAST"
    },
    {
        "offset": -2.5,
        "name": "Heure Avancée de Terre-Neuve",
        "code": "HAT"
    },
    {
        "offset": -8.0,
        "name": "Heure Avancée du Yukon",
        "code": "HAY"
    },
    {
        "offset": 8.0,
        "name": "Hong Kong Time",
        "code": "HKT"
    },
    {
        "offset": -4.5,
        "name": "Hora Legal de Venezuela",
        "code": "HLV"
    },
    {
        "offset": -4.0,
        "name": "Heure Normale de l'Atlantique",
        "code": "HNA"
    },
    {
        "offset": -6.0,
        "name": "Heure Normale du Centre",
        "code": "HNC"
    },
    {
        "offset": -5.0,
        "name": "Heure Normale de l'Est",
        "code": "HNE"
    },
    {
        "offset": -8.0,
        "name": "Heure Normale du Pacifique",
        "code": "HNP"
    },
    {
        "offset": -7.0,
        "name": "Heure Normale des Rocheuses",
        "code": "HNR"
    },
    {
        "offset": -3.5,
        "name": "Heure Normale de Terre-Neuve",
        "code": "HNT"
    },
    {
        "offset": -9.0,
        "name": "Heure Normale du Yukon",
        "code": "HNY"
    },
    {
        "offset": 7.0,
        "name": "Hovd Time",
        "code": "HOVT"
    },
    {
        "offset": 7.0,
        "name": "Indochina Time",
        "code": "ICT"
    },
    {
        "offset": 3.0,
        "name": "Israel Daylight Time",
        "code": "IDT"
    },
    {
        "offset": 6.0,
        "name": "Indian Chagos Time",
        "code": "IOT"
    },
    {
        "offset": 4.5,
        "name": "Iran Daylight Time",
        "code": "IRDT"
    },
    {
        "offset": 9.0,
        "name": "Irkutsk Summer Time",
        "code": "IRKST"
    },
    {
        "offset": 9.0,
        "name": "Irkutsk Time",
        "code": "IRKT"
    },
    {
        "offset": 3.5,
        "name": "Iran Standard Time",
        "code": "IRST"
    },
    {
        "offset": 2.0,
        "name": "Israel Standard Time",
        "code": "UNUSED_IST"
    },
    {
        "offset": 5.5,
        "name": "India Standard Time",
        "code": "IST"
    },
    {
        "offset": 1.0,
        "name": "Irish Standard Time",
        "code": "IST"
    },
    {
        "offset": 9.0,
        "name": "Japan Standard Time",
        "code": "JST"
    },
    {
        "offset": 6.0,
        "name": "Kyrgyzstan Time",
        "code": "KGT"
    },
    {
        "offset": 8.0,
        "name": "Krasnoyarsk Summer Time",
        "code": "KRAST"
    },
    {
        "offset": 8.0,
        "name": "Krasnoyarsk Time",
        "code": "KRAT"
    },
    {
        "offset": 9.0,
        "name": "Korea Standard Time",
        "code": "KST"
    },
    {
        "offset": 4.0,
        "name": "Kuybyshev Time",
        "code": "KUYT"
    },
    {
        "offset": 11.0,
        "name": "Lord Howe Daylight Time",
        "code": "LHDT"
    },
    {
        "offset": 10.5,
        "name": "Lord Howe Standard Time",
        "code": "LHST"
    },
    {
        "offset": 14.0,
        "name": "Line Islands Time",
        "code": "LINT"
    },
    {
        "offset": 12.0,
        "name": "Magadan Summer Time",
        "code": "MAGST"
    },
    {
        "offset": 12.0,
        "name": "Magadan Time",
        "code": "MAGT"
    },
    {
        "offset": -9.5,
        "name": "Marquesas Time",
        "code": "MART"
    },
    {
        "offset": 5.0,
        "name": "Mawson Time",
        "code": "MAWT"
    },
    {
        "offset": -6.0,
        "name": "Mountain Daylight Time",
        "code": "MDT"
    },
    {
        "offset": 2.0,
        "name": "Mitteleuropäische Sommerzeit",
        "code": "MESZ"
    },
    {
        "offset": 1.0,
        "name": "Mitteleuropäische Zeit",
        "code": "MEZ"
    },
    {
        "offset": 12.0,
        "name": "Marshall Islands Time",
        "code": "MHT"
    },
    {
        "offset": 6.5,
        "name": "Myanmar Time",
        "code": "MMT"
    },
    {
        "offset": 4.0,
        "name": "Moscow Daylight Time",
        "code": "MSD"
    },
    {
        "offset": 4.0,
        "name": "Moscow Standard Time",
        "code": "MSK"
    },
    {
        "offset": -7.0,
        "name": "Mountain Standard Time",
        "code": "MST"
    },
    {
        "offset": 4.0,
        "name": "Mauritius Time",
        "code": "MUT"
    },
    {
        "offset": 5.0,
        "name": "Maldives Time",
        "code": "MVT"
    },
    {
        "offset": 8.0,
        "name": "Malaysia Time",
        "code": "MYT"
    },
    {
        "offset": 11.0,
        "name": "New Caledonia Time",
        "code": "NCT"
    },
    {
        "offset": -2.5,
        "name": "Newfoundland Daylight Time",
        "code": "NDT"
    },
    {
        "offset": 11.5,
        "name": "Norfolk Time",
        "code": "NFT"
    },
    {
        "offset": 7.0,
        "name": "Novosibirsk Summer Time",
        "code": "NOVST"
    },
    {
        "offset": 6.0,
        "name": "Novosibirsk Time",
        "code": "NOVT"
    },
    {
        "offset": 5.75,
        "name": "Nepal Time",
        "code": "NPT"
    },
    {
        "offset": -3.5,
        "name": "Newfoundland Standard Time",
        "code": "NST"
    },
    {
        "offset": -11.0,
        "name": "Niue Time",
        "code": "NUT"
    },
    {
        "offset": 13.0,
        "name": "New Zealand Daylight Time",
        "code": "NZDT"
    },
    {
        "offset": 12.0,
        "name": "New Zealand Standard Time",
        "code": "NZST"
    },
    {
        "offset": 7.0,
        "name": "Omsk Summer Time",
        "code": "OMSST"
    },
    {
        "offset": 7.0,
        "name": "Omsk Standard Time",
        "code": "OMST"
    },
    {
        "offset": -7.0,
        "name": "Pacific Daylight Time",
        "code": "PDT"
    },
    {
        "offset": -5.0,
        "name": "Peru Time",
        "code": "PET"
    },
    {
        "offset": 12.0,
        "name": "Kamchatka Summer Time",
        "code": "PETST"
    },
    {
        "offset": 12.0,
        "name": "Kamchatka Time",
        "code": "PETT"
    },
    {
        "offset": 10.0,
        "name": "Papua New Guinea Time",
        "code": "PGT"
    },
    {
        "offset": 13.0,
        "name": "Phoenix Island Time",
        "code": "PHOT"
    },
    {
        "offset": 8.0,
        "name": "Philippine Time",
        "code": "PHT"
    },
    {
        "offset": 5.0,
        "name": "Pakistan Standard Time",
        "code": "PKT"
    },
    {
        "offset": -2.0,
        "name": "Pierre & Miquelon Daylight Time",
        "code": "PMDT"
    },
    {
        "offset": -3.0,
        "name": "Pierre & Miquelon Standard Time",
        "code": "PMST"
    },
    {
        "offset": 11.0,
        "name": "Pohnpei Standard Time",
        "code": "PONT"
    },
    {
        "offset": -8.0,
        "name": "Pacific Standard Time",
        "code": "PST"
    },
    {
        "offset": -8.0,
        "name": "Pitcairn Standard Time",
        "code": "UNUSED_PST"
    },
    {
        "offset": 9.0,
        "name": "Palau Time",
        "code": "PWT"
    },
    {
        "offset": -3.0,
        "name": "Paraguay Summer Time",
        "code": "PYST"
    },
    {
        "offset": -4.0,
        "name": "Paraguay Time",
        "code": "PYT"
    },
    {
        "offset": 4.0,
        "name": "Reunion Time",
        "code": "RET"
    },
    {
        "offset": 4.0,
        "name": "Samara Time",
        "code": "SAMT"
    },
    {
        "offset": 2.0,
        "name": "South Africa Standard Time",
        "code": "SAST"
    },
    {
        "offset": 11.0,
        "name": "Solomon IslandsTime",
        "code": "SBT"
    },
    {
        "offset": 4.0,
        "name": "Seychelles Time",
        "code": "SCT"
    },
    {
        "offset": 8.0,
        "name": "Singapore Time",
        "code": "SGT"
    },
    {
        "offset": -3.0,
        "name": "Suriname Time",
        "code": "SRT"
    },
    {
        "offset": -11.0,
        "name": "Samoa Standard Time",
        "code": "SST"
    },
    {
        "offset": -10.0,
        "name": "Tahiti Time",
        "code": "TAHT"
    },
    {
        "offset": 5.0,
        "name": "French Southern and Antarctic Time",
        "code": "TFT"
    },
    {
        "offset": 5.0,
        "name": "Tajikistan Time",
        "code": "TJT"
    },
    {
        "offset": 13.0,
        "name": "Tokelau Time",
        "code": "TKT"
    },
    {
        "offset": 9.0,
        "name": "East Timor Time",
        "code": "TLT"
    },
    {
        "offset": 5.0,
        "name": "Turkmenistan Time",
        "code": "TMT"
    },
    {
        "offset": 12.0,
        "name": "Tuvalu Time",
        "code": "TVT"
    },
    {
        "offset": 8.0,
        "name": "Ulaanbaatar Time",
        "code": "ULAT"
    },
    {
        "offset": 0.0,
        "name": "Coordinated Universal Time",
        "code": "UTC"
    },
    {
        "offset": -2.0,
        "name": "Uruguay Summer Time",
        "code": "UYST"
    },
    {
        "offset": -3.0,
        "name": "Uruguay Time",
        "code": "UYT"
    },
    {
        "offset": 5.0,
        "name": "Uzbekistan Time",
        "code": "UZT"
    },
    {
        "offset": -4.5,
        "name": "Venezuelan Standard Time",
        "code": "VET"
    },
    {
        "offset": 11.0,
        "name": "Vladivostok Summer Time",
        "code": "VLAST"
    },
    {
        "offset": 11.0,
        "name": "Vladivostok Time",
        "code": "VLAT"
    },
    {
        "offset": 11.0,
        "name": "Vanuatu Time",
        "code": "VUT"
    },
    {
        "offset": 2.0,
        "name": "West Africa Summer Time",
        "code": "WAST"
    },
    {
        "offset": 1.0,
        "name": "West Africa Time",
        "code": "WAT"
    },
    {
        "offset": 1.0,
        "name": "Western European Summer Time",
        "code": "WEST"
    },
    {
        "offset": 1.0,
        "name": "Westeuropäische Sommerzeit",
        "code": "WESZ"
    },
    {
        "offset": 0.0,
        "name": "Western European Time",
        "code": "WET"
    },
    {
        "offset": 0.0,
        "name": "Westeuropäische Zeit",
        "code": "WEZ"
    },
    {
        "offset": 12.0,
        "name": "Wallis and Futuna Time",
        "code": "WFT"
    },
    {
        "offset": -2.0,
        "name": "Western Greenland Summer Time",
        "code": "WGST"
    },
    {
        "offset": -3.0,
        "name": "West Greenland Time",
        "code": "WGT"
    },
    {
        "offset": 7.0,
        "name": "Western Indonesian Time",
        "code": "WIB"
    },
    {
        "offset": 9.0,
        "name": "Eastern Indonesian Time",
        "code": "WIT"
    },
    {
        "offset": 8.0,
        "name": "Central Indonesian Time",
        "code": "WITA"
    },
    {
        "offset": 1.0,
        "name": "Western Sahara Summer Time",
        "code": "UNUSED_WST"
    },
    {
        "offset": 13.0,
        "name": "West Samoa Time",
        "code": "WST"
    },
    {
        "offset": 10.0,
        "name": "Yakutsk Summer Time",
        "code": "YAKST"
    },
    {
        "offset": 10.0,
        "name": "Yakutsk Time",
        "code": "YAKT"
    },
    {
        "offset": 10.0,
        "name": "Yap Time",
        "code": "YAPT"
    },
    {
        "offset": 6.0,
        "name": "Yekaterinburg Summer Time",
        "code": "YEKST"
    },
    {
        "offset": 6.0,
        "name": " Yekaterinburg Time",
        "code": "YEKT"
    }
]



import re


def join(x, y=None):
    if y is None:
        return "|".join(x)
    else:
        r = y + "|"
        return r.join(x) + y


def group(x):
    return "({})".format(x)

RE_UNITS = group(join(UNITS, "s?"))
RE_ORDINAL = group(join(ORDINAL_NUMBERS))
RE_LAST_FIRST = group(join(LAST_FIRST))
SS = "[ -]+"

t = "within the last fourty 4 weeks"
re.search("within the " + RE_LAST_FIRST + SS + group(RE_ORDINAL) + "+" + RE_UNITS, t).groups()

re.search("within the " + RE_LAST_FIRST + SS + RE_ORDINAL + RE_ORDINAL + RE_UNITS, t).groups()

matches = []
matches += match(RE_LAST_FIRST, t)
matches += match(RE_ORDINAL, t)
matches += match(RE_UNITS, t)
matches += match("\d+", t)


def match(needle, haystack):
    return [(x.string[x.span()[0]:x.span()[1]], x.span()) for x in re.finditer(needle, haystack)]


haha = "(" + ")|(".join(ORDINAL_NUMBERS) + ")"
RE_ORDINAL = "(?:(?:" + haha + ")[ -]+)+"

t = "within the last fourty four weeks"
list(filter(None, re.search("within the " + RE_LAST_FIRST + SS + RE_ORDINAL + RE_UNITS, t).groups()))


re.findall(RE_ORDINAL + r"[ -]+(?=.*\bitems)", s)



{
    "3 people at most 500 dollars june 2018, to a very hot destination": {},
    "The 3 of us around $100 at 10:00 on 25th of June to Brussels.": {},
    "The 3 of us around $100 at 10:00 on 25th of June in 2018 to Brussels.": {},
    " For 100 euros I want to go to a skiing resort this year.": {},
    " 5 people went on holiday to Spain in August.": {},
    "I want to fly to Paris for at least 500 dollars in 2017": {},
    "3 people at most 500 dollars 04-06-2017": {},
    "Aug 2": {},
    "Janeiro": {},
    "this winter": {},
    "this Summer": {},
    "Next fall": {},
    "in the next 10 minutes i will leave to france": {},
    "this year": {},
    "next year": {},
    "leave today": {},
    "tomorrow": {},
    "leave tomorrow": {},
    "next week": {},
    "you are week": {},
    "in summer 2017": {},
    "in June summer 2017": {},
    "Summer of '69": {},
    "2017-06-25 00:00:00": {},
    "2017-06-25 00:00:00.123123": {},
    "2017-06-25 00:01:01": {},
    "00:01:01 2017-06-25": {},
    "2015-06-07": {},
    "3:00pm": {},
    "3:00 pm": {},
    "12:00 pm": {},
    "3 pm": {},
    "three pm": {},
    "3.00pm": {},
    "a year earlier": {},
    "in 3 weeks and 5 days": {},
    "The 3 of us will each buy 2 cars 2 weeks after 5th of June in 2018 at 10:05:01 for 100 dollars.": {},
    "The 3 of us will each buy 2 cars 2 weeks after 5th of June in 2018 for 100 dollars.": {},
    "some text (March 9, 1990) more text": {},
    "some text (March 9th, 1990) more text": {},
    "March 9, 1990": {},
    "1 week and one day ago": {},
    "friday in two weeks": {},
    "for 2 weeks": {},
    "for 2 weeks and 5 days": {},
    "previous monday": {},
    "Tue two weeks ago": {},
    "last wednesday": {},
    "last wednesday last June": {},
    "in the first 10 days of June": {},
    "over the next five days": {},
    "over the next twenty 4 years": {},
    "during the first five and a half days of June": {},
    "over the last five and a half days of June": {},
    "from Saturday to Thursday inclusive": {},
    "from a year earlier": {},
    "in the first 20 days": {},
    "in the last 20 days of February": {},
    "in the last few days": {},
    "Three interesting dates are 18 Feb 1997": {},
    "next tuesday at 11pm": {},
    "6 days 23 hours 58 minutes": {},
    "between June 16 and 18": {},
    "between June and August at 10pm": {},
    "between June and August at 10:00 p.m.": {},
    "between June and August at ten p.m.": {},
    "today for two days": {},
    "between sunday and saturday": {},
    "during the last 10 days of March": {},
    "last year": {},
    "last quarter": {},
    "next quarter": {},
    "during the first 78 days of the year": {},
    "Three interesting dates are 18 Feb 1997, the 20th of july and 4 days from today.": {},
    "over the first 7 days in 1 month": {},
    "3 months starting next month": {},
    "in the first 2 weeks of next month": {},
    "last 2 weeks next june": {},
    "on the thirty first of june": {},
    "three in the afternoon": {},
    "afternoon": {},
    "in the year nineteen seventy": {},
    "Q1": {},
    "this quarter": {},
    "i will learn french within 3 weeks": {},
    "i will learn french after 3 weeks": {},
    "every 2 weeks between this year and the next until November on Tuesdays and Wednesdays at ten and 11 pm for 10 minutes": {},
    "two weeks tomorrow": {},
    "two weeks ago tomorrow": {}
}



import re
from itertools import product
from datetime import datetime

from metadate.classes import MetaRelative
from metadate.classes import MetaOrdinal
from metadate.classes import MetaUnit
from metadate.classes import MetaRange
from metadate.classes import MetaBetween
from metadate.classes import MetaEvery
from metadate.classes import MetaAnd
from metadate.classes import MetaModifier
from metadate.classes import MetaDuration
from metadate.utils import strip_pm
from metadate.utils import Units

RYEAR = ["[12][0-9]{3}"]
RMONTH = ["0?[1-9]", "1[0-2]"]
RDAY = ['[12][0-9]', '[3][01]', '0[1-9]', '[0-9]']
SEP = ['[ -/]']
RHOUR = ['[01][0-9]', '2[0-3]', '[0-9]']
RMINUTE = ['[0-5][0-9]']
RSECOND = ['[0-5][0-9]']
END = [r'\b']
APM = [r" ?a\.?m\.?", r" ?p\.?m\.?", r" afternoon", r".?o'?clock", "'"]


def pipe(ls, pre=r'\b', post=r''):
    # this is flipped
    p = post + "|" + pre
    return pre + p.join(sorted(ls, key=lambda x: len(x), reverse=True)) + post


def scan_product(*args):
    return '|'.join([''.join(x) for x in product(*args)])


def validate(second=None, minute=None, hour=None, day=None, month=None, year=None):
    if second is not None and second not in range(0, 60):
        return False
    if minute is not None and minute not in range(0, 60):
        return False
    if hour is not None and hour not in range(0, 24):
        return False
    if day is not None and day not in range(1, 32):
        return False
    if month is not None and month not in range(1, 13):
        return False
    if year is not None and year not in range(1900, 2100):
        return False
    return True


def day_month_year_switch_validate(day, month, year, dd_left_first):
    if not dd_left_first:
        day, month = month, day
    if validate(day=day, month=month, year=year):
        return True, day, month
    day, month = month, day
    if validate(day=day, month=month, year=year):
        return True, day, month
    return False, None, None


class Scanner():

    def __init__(self, locale, re_flags=re.IGNORECASE | re.MULTILINE):
        self.__dict__.update({k: v for k, v in locale.__dict__.items() if not k.startswith("__")})

        self.blacklist = []
        for k, v in locale.__dict__.items():
            if isinstance(v, (dict, list)) and k == k.upper():
                self.blacklist.extend(list(v))

        self.HH_MM_SS = scan_product(END, RHOUR, ["[:h]"], RMINUTE, ["[:m]"],
                                     RSECOND, APM + ["m?"], END)
        # probably not so needed
        self.HH_MM_SS_m = scan_product(END, RHOUR, ["[:h]"], RMINUTE, ["[:m]"],
                                       RSECOND, ['\.\d+'], END)

        self.HH_MM = scan_product(END, RHOUR, ["[:h']"], RMINUTE, APM + ["[hm]?"], END)
        #self.HH_MM2 = scan_product(END, RHOUR, ["."], RMINUTE, APM, END)
        self.HHAPM = scan_product(END, RHOUR, APM, END)
        self.YYYY_MM_DD = scan_product(END, RYEAR, SEP, RMONTH, SEP, RDAY, END)
        self.YYYYMMDD = scan_product(END, RYEAR, RMONTH, RDAY, END)
        self.DD_MM_YYYY = scan_product(END, RDAY, SEP, RMONTH, SEP, RYEAR, END)
        self.AMBIGUOUS_DD_DD_YYYY = scan_product(
            END, RDAY, ["[./ -]"], RDAY, ["[./ -]"], RYEAR + ["[01289][0-9]"], END)
        self.DDMMYYYY = scan_product(END, RDAY, RMONTH, RYEAR, END)
        self.YYYY = scan_product(END, RYEAR, END)
        self.AMBIGUOUS_DD_DD = scan_product(
            ["on ", "the ", "for ", "from ", "until "], RDAY, ["[./ -]"], RDAY, END, ["(?![./0-9-][./0-9-])"])
        self.DDMM = scan_product(END, [r"\d{1,2}"], [" ?..", ""], [
                                 " of ", " ", "-", ""], self.MONTHS)
        self.DDMMS = scan_product(END, [r"\d{1,2}"], [" ?[snrt][tdh]", ""], [
            " of ", " ", "-", ""], self.MONTHS_SHORTS)
        self.LETTER_MONTH_DAY = scan_product(END, self.MONTHS, [" [0-3]?[0-9]"],
                                             [" ?"], self.RANK_NAMING + [""], END)
        self.SHORT_DAY_MONTH = scan_product(END, self.MONTHS_SHORTS, ["[.]?"], ["[ -][0-3]?[0-9]"],
                                            [" ?"], self.RANK_NAMING + [""], END)
        self.ORDINAL_APM = scan_product(END, self.ORDINAL_NUMBERS, [" a\.?m", " p\.?m"])

        self.ORDINAL = scan_product(END, self.ORDINAL_NUMBERS, [" "])
        self.ON_THE_DAY = scan_product(END, self.ON_THE, RDAY, [" ?"], self.RANK_NAMING, END)
        self.AT_HOUR = scan_product(END, self.AT, [" "], RHOUR, ["h?"], ["(?![:']|[.] )"], END)

        # IN_THE = ['in the', 'over the', 'during the']
        # LAST_FIRST = {
        #     'last': -1,
        #     'first': 1
        # }

        # MODIFIERS = {
        #     "in": 1,
        #     "on": 0,
        #     "this": 0,
        #     "next":  1,
        #     "coming": 1
        # }

        self.scanner = re.Scanner([
            (self.HH_MM_SS_m, self.hh_mm_ss),
            (self.HH_MM_SS, self.hh_mm_ss),
            (self.HH_MM, self.hh_mm),
            #(self.HH_MM2, self.hh_mm),
            (self.HHAPM, self.hh_mm),
            (self.YYYY_MM_DD, self.yyyy_mm_dd),
            (self.YYYYMMDD, self.yyyymmdd),
            # these ambigious have a bad performance
            #(self.AMBIGUOUS_DD_DD_YYYY, self.ambiguous_dd_dd_yyyy),
            #(self.AMBIGUOUS_DD_DD, self.ambiguous_dd_dd),
            (self.DD_MM_YYYY, self.dd_mm_yyyy),
            #(self.DDMMYYYY, self.ddmmyyyy),
            (self.YYYY, self.yyyy),                      # YYYY
            (self.ORDINAL_APM, self.ordinal_apm),
            (self.ORDINAL, self.ordinal),
            #(pipe(self.NOW, post=r"\b"), self.now),
            #(pipe(self.BETWEEN, post=r'\b'), lambda y, x: MetaBetween(None, None, None, span=y.match.span())),
            #(pipe(self.EVERY, post=r'\b'), self.every),
            (pipe(self.IN_THE, post=" "), self.in_the),
            #(pipe(self.DURATION, post=" "), self.duration),
            (pipe(self.TODAY_TOMORROW), self.today_tomorrow),
            #(pipe(self.SEASONS, post=r'\b'), self.season),
            (pipe(self.QUARTERS, post=r'\b'), self.quarter),
            (pipe(self.MODIFIERS, post=r'\b'), self.modifier),
            (self.LETTER_MONTH_DAY, self.letter_month_day),
            (self.SHORT_DAY_MONTH, self.short_day_month),
            (self.DDMM, self.ddmm),
            (self.DDMMS, self.ddmms),
            (pipe(self.MONTHS, post=r'\b(?! i)'), self.letter_month),
            (pipe(self.MONTHS_SHORTS, post=r'[.]?\b(?! i\b)'), self.short_month),
            (pipe(self.WEEKDAY, post=r'\b'), self.weekday),
            (pipe(self.WEEKDAY_SHORTS, post=r'\b'), self.weekday_shorts),
            (self.ON_THE_DAY, self.on_the_day),
            (self.AT_HOUR, self.at_hour),
            ("and a half", lambda y, x: MetaOrdinal("0.5", span=y.match.span())),
            # (self.ON_THE_DAY, self.on_the_day_ordinal),
            (pipe(self.UNITS, r'\b', r'\b'), lambda y, x: MetaUnit(
                self.UNITS[x.lower()], span=y.match.span())),
            (r"\d+[.,]?[0-9]?", lambda y, x: MetaOrdinal(x.replace(",", "."), span=y.match.span())),

            # temps
            ("[.][.]+", "ellipsis"),
            (r"\b[a-zA-Z]{1,3}[.](?![a-zA-Z0-9])[^.]", "abbrev"),
            ("[!?]+(?![a-zA-Z0-9]) ?", "SENT"),
            ("[.](?![a-zA-Z0-9.]) ?", "SENT"),
            ("\n", "SENT"),
            # tricky stuff
            (pipe(self.AND), lambda y, x: MetaAnd(x, span=y.match.span())),
            (pipe(self.blacklist), None),
            (r' +', None),
            (r'-', None),  # can be "5-6 days" and "2009-2010"
            (r'.', lambda y, x: x)
        ], re_flags)

    def scan(self, text):
        text = text.replace("–", "-")
        return self.scanner.scan(text)

    def season(self, y, x):
        return MetaRelative(month=self.SEASONS[x.lower()], day=21, levels=set([Units.SEASON]),
                            span=y.match.span())

    def quarter(self, y, x):
        return MetaRelative(month=self.QUARTERS[x.lower()], day=1, levels=set([Units.QUARTER]),
                            span=y.match.span())

    def modifier(self, y, x):
        x = x.lower()
        return MetaModifier(x, self.MODIFIERS[x], span=y.match.span())

    def in_the(self, y, x):
        s, e = y.match.span()
        e -= 1
        return MetaRange(x.rstrip(" "), span=(s, e))

    @staticmethod
    def duration(y, x):
        s, e = y.match.span()
        e -= 1
        return MetaDuration(None, None, span=(s, e))

    @staticmethod
    def every(y, x):
        s, e = y.match.span()
        e -= 1
        return MetaEvery(None, None, None, span=(s, e))

    def letter_month_day(self, y, x):
        # June 16
        month, day = re.sub("[a-zA-Z]+$", "", x).lower().split()
        return MetaRelative(month=self.MONTHS[month], day=int(day), span=y.match.span())

    def letter_day_month(self, y, x):
        # 16 June
        day, month = x.lower().split()
        return MetaRelative(month=self.MONTHS[month], day=int(day), span=y.match.span())

    def short_month_day(self, y, x):
        month, day = re.sub("[a-zA-Z.]+$", "", x).lower().split()
        return MetaRelative(month=self.MONTHS_SHORTS[month], day=int(day), span=y.match.span())

    def short_day_month(self, y, x):
        month, day = re.split("[^a-zA-Z0-9]+", x.lower().strip())
        day = re.sub("[stndrh]", "", day)
        return MetaRelative(month=self.MONTHS_SHORTS[month[:3]], day=int(day), span=y.match.span())

    def weekday(self, y, x):
        # Tuesday
        x = x.replace("-", " ")
        weekday = self.WEEKDAY[x.lower().rstrip("s")]
        return MetaRelative(weekday=weekday, span=y.match.span())

    def weekday_shorts(self, y, x):
        # Tuesday
        x = x.replace("-", " ")
        weekday = self.WEEKDAY_SHORTS[x.lower().rstrip("s")]
        return MetaRelative(weekday=weekday, span=y.match.span())

    def ddmm(self, y, x):
        x = x.lower()
        month = re.findall("|".join(self.MONTHS), x)[0]
        day = re.sub("[^0-9]", "", x)
        return MetaRelative(month=self.MONTHS[month], day=int(day), span=y.match.span())

    def ddmms(self, y, x):
        x = x.replace("-", " ")
        # now to cover 31mar
        parts = x[:-3], x[-4:]
        day = re.sub("[^0-9]", "", parts[0])
        month = re.sub("[0-9]", "", parts[-1].lower().strip())
        return MetaRelative(month=self.MONTHS_SHORTS[month], day=int(day), span=y.match.span())

    def letter_month(self, y, x):
        return MetaRelative(month=self.MONTHS[x.lower()], span=y.match.span())

    def short_month(self, y, x):
        return MetaRelative(month=self.MONTHS_SHORTS[x[:3].lower()], span=y.match.span())

    def today_tomorrow(self, y, x):
        days = self.TODAY_TOMORROW[x.lower()]
        return MetaRelative(days=days, span=y.match.span())

    def th_of_month(self, y, x):
        # 25th of June
        x = x.replace("-", " ")
        parts = x.lower().split()
        dayth, month = parts[0][:-2], parts[-1]
        return MetaRelative(month=self.MONTHS[month], day=int(dayth), span=y.match.span())

    @staticmethod
    def now(y, x):
        now = datetime.now()
        if len(y.match.string) > 5:
            # print("skipping now")
            return None
        return MetaRelative(year=now.year, month=now.month, day=now.day, hour=now.hour,
                            minute=now.minute, second=now.second, span=y.match.span())

    @staticmethod
    def hh_mm_ss(y, x):
        hour, minute, second, microsecond = strip_pm(x)
        if "." in x:
            md = MetaRelative(hour=hour, minute=minute, second=second,
                              microsecond=microsecond, span=y.match.span())
        else:
            md = MetaRelative(hour=hour, minute=minute, second=second,
                              span=y.match.span())
        return md

    @staticmethod
    def hh_mm(y, x):
        hour, minute, _, _ = strip_pm(x)
        if minute is None:
            return MetaRelative(hour=hour, span=y.match.span())
        return MetaRelative(hour=hour, minute=minute, span=y.match.span())

    def ordinal_apm(self, y, x):
        hour, minute, second, _ = strip_pm(x, numbers_dict=self.ORDINAL_NUMBERS)
        return MetaRelative(hour=hour, minute=minute, second=second, levels=set([Units.MINUTE, Units.HOUR]),
                            span=y.match.span())

    def ordinal(self, y, x):
        return MetaOrdinal(self.ORDINAL_NUMBERS[x.lower().strip().replace(",", ".")], span=y.match.span())

    @staticmethod
    def yyyy_mm_dd(y, x):
        parts = re.split("[ /-]", x)
        if len(parts) != 3:
            return None
        year, month, day = parts
        return MetaRelative(year=int(year), month=int(month), day=int(day), span=y.match.span())

    @staticmethod
    def yyyymmdd(y, x):
        if len(x) != 8:
            return None
        year, month, day = int(x[:4]), int(x[4:6]), int(x[6:8])
        return MetaRelative(year=year, month=month, day=day, span=y.match.span())

    @staticmethod
    def dd_mm_yyyy(y, x):
        parts = re.split("[ /.-]", x)
        if len(parts) != 3:
            return None
        day, month, year = parts
        return MetaRelative(year=int(year), month=int(month), day=int(day), span=y.match.span())

    @staticmethod
    def ddmmyyyy(y, x):
        # TODO: 1 01 2015 / 01 1 2015, 1 and 2
        if len(x) == 6:
            return None
        else:
            year, month, day = int(x[4:8]), int(x[2:4]), int(x[:2])
        return MetaRelative(year=year, month=month, day=day, span=y.match.span())

    def ambiguous_dd_dd_yyyy(self, y, x):
        seps = re.findall("[ ./-]", x)
        num_sep = len(seps)
        num_numeric = len(x) - num_sep
        # different separators is not allowed
        # only 1 sep is also not allowed
        if len(set(seps)) > 1:
            return None
        if num_numeric == 8 and num_sep != 1:
            x = re.sub("[ ./-]", "", x)
            day, month, year = x[:2], x[2:4], x[4:8]
            day, month, year = int(day), int(month), int(year)
            valid, day, month = day_month_year_switch_validate(
                day, month, year, self.DD_LEFT_FIRST)
            if not valid:
                return None
        elif num_numeric == 7 and num_sep == 2:
            day, month, year = x.split(seps[0])
            day, month, year = int(day), int(month), int(year)
            valid, day, month = day_month_year_switch_validate(
                day, month, year, self.DD_LEFT_FIRST)
            if not valid:
                return None
        # 1-1-2016, dayfirst check
        elif num_numeric == 6 and num_sep == 2:
            sep = seps[0]
            day, month, year = x.split(sep)
            day, month, year = int(day), int(month), int(year)
            valid, day, month = day_month_year_switch_validate(
                day, month, year, self.DD_LEFT_FIRST)
            if not valid:
                return None
        # 02-2010 no dayfirst needed, has to be month
        elif num_numeric == 6 and num_sep == 1:
            sep = seps[0]
            month, year = x.split(sep)
            day, month = int(day), int(month)
            if not validate(year=year, month=month):
                return None
            else:
                return MetaRelative(year=year, month=month, span=y.match.span())
        # elif num_numeric == 5 and num_sep == 1:
        # # 1-1-16 and 7-11-17
        # dayfirst check
        elif num_numeric in (4, 5) and num_sep == 2:
            sep = seps[0]
            day, month, year = x.split(sep)
            day, month = int(day), int(month)
            year = int("20" + str(year))
            valid, day, month = day_month_year_switch_validate(
                day, month, year, self.DD_LEFT_FIRST)
            if not valid:
                return None
        else:
            raise ValueError("Cannot understand '{}'".format(x))
            # let's remind ourselves that this is not going to matter in some years
        return MetaRelative(year=year, month=month, day=day, span=y.match.span())

    @staticmethod
    def ambiguous_dd_dd(y, x):
        _, day, month = re.split("[ ./-]", x)
        day, month = int(day), int(month)
        if not validate(day=day, month=month):
            if validate(day=month, month=day):
                day, month = month, day
            else:
                return None
        return MetaRelative(month=month, day=day, span=y.match.span())

    @staticmethod
    def yyyy(y, x):
        return MetaRelative(year=int(x), span=y.match.span())

    @staticmethod
    def on_the_day(y, x):
        # "on the 31st"
        return MetaRelative(day=int(x.split()[2][:-2]), span=y.match.span())

    @staticmethod
    def at_hour(y, x):
        # "at 19"
        x = x.strip("h")
        return MetaRelative(hour=int(x.split()[1]), span=y.match.span())



import sys
from metadate import parse_date


def main():
    """ This is the function that is run from commandline with `metadate` """
    if "--lang" == sys.argv[1]:
        dates = parse_date(" ".join(sys.argv[3:]), lang=sys.argv[2], multi=True)
    else:
        dates = parse_date(" ".join(sys.argv[1:]), multi=True)
    if len(dates) == 2:
        print(dates[1].start_date - dates[0].start_date)
    else:
        print(dates)



import re
import collections
import datetime as DT
from dateutil.tz import tzoffset, tzutc
import pytz
from metadate.utils import Units

abbrevs = set(["EDT"])

for name in pytz.all_timezones:
    tzone = pytz.timezone(name)
    for utcoffset, dstoffset, tzabbrev in getattr(
        tzone, '_transition_info', [[None, None, DT.datetime.now(tzone).tzname()]]
    ):
        abbrevs.add(name)

TIMEZONE_3_RE = re.compile("|".join(sorted([x for x in abbrevs if x == x.upper()])))


def timezone_1(mp):
    beg = mp.spans[0][0]
    end = mp.spans[-1][1]
    match = re.search("([+-])(0[0-9]|1[012])([0-9][0-9])", mp.text[beg : end + 10])
    if match is not None:
        direction, hour, minute = match.groups()
        offset = int(hour) * 3600 + int(minute) * 60
        direction = 1 if direction == "+" else -1
        offset *= direction
        if offset == 0:
            tz = tzutc()
        else:
            tz = tzoffset(None, offset)
        span = match.span()
        return span, tz
    return None, None


def timezone_2(mp):
    beg = mp.spans[0][0]
    end = mp.spans[-1][1]
    match = re.search(r"(GMT|UTC)([+-])([0-9]|1[012])\b", mp.text[beg : end + 10])
    if match is not None:
        _, direction, hour = match.groups()
        offset = int(hour) * 3600
        direction = 1 if direction == "+" else -1
        offset *= direction
        if offset == 0:
            tz = tzutc()
        else:
            tz = tzoffset(None, offset)
        span = match.span()
        return span, tz
    return None, None


def timezone_3(mp):
    beg = mp.spans[0][0]
    end = mp.spans[-1][1]
    match = TIMEZONE_3_RE.search(mp.text[beg : end + 10])
    if match is not None:
        name = match.group().replace("EDT", "EST")
        tz = pytz.timezone(name)
        span = match.span()
        return span, tz
    return None, None


def add_timezone(mp):
    if mp.min_level > Units.HOUR:
        return mp
    beg = mp.spans[0][0]
    span, tz = timezone_1(mp)
    if span is None:
        span, tz = timezone_2(mp)
    if span is None:
        span, tz = timezone_3(mp)
    if span is None:
        return mp
    mp.start_date = mp.start_date.replace(tzinfo=tz)
    mp.end_date = mp.end_date.replace(tzinfo=tz)
    mp.matches.append(mp.text[span[0] + beg : span[1] + beg])
    mp.spans.append((span[0] + beg, span[1] + beg))
    mp.spans = sorted(mp.spans)
    return mp



import re
from enum import IntEnum

from dateutil.relativedelta import relativedelta as rd


class Units(IntEnum):
    YEAR = 9
    SEASON = 8
    QUARTER = 7
    MONTH = 6
    WEEK = 5
    DAY = 4
    HOUR = 3
    MINUTE = 2
    SECOND = 1
    MICROSECOND = 0


RD_ARG_TO_UNIT = {
    "years": Units.YEAR,
    "year": Units.YEAR,
    "months": Units.MONTH,
    "month": Units.MONTH,
    "weeks": Units.WEEK,
    "week": Units.WEEK,
    "weekday": Units.DAY,
    "days": Units.DAY,
    "day": Units.DAY,
    "hours": Units.HOUR,
    "hour": Units.HOUR,
    "minutes": Units.MINUTE,
    "minute": Units.MINUTE,
    "seconds": Units.SECOND,
    "second": Units.SECOND,
    "microseconds": Units.MICROSECOND,
    "microsecond": Units.MICROSECOND,
}

freqs = ["yearly", "monthly", "weekly", "daily", "hourly", "minutely", "secondly"]
FREQ = {x: num for num, x in enumerate(freqs)}


def log(tag, x, verbose=False):
    if verbose:
        print("--- {} ------------".format(tag))
        print(x)


def add_tag(sentence, matches, color="mediumspringgreen"):
    # given textual matches between ranges like [(5, 10), (10, 15)]
    # this will clean up
    # first is 0,5
    news = ''
    lbound = 0
    hit = "<span style='background-color:{}'>{}</span>"
    for m in matches:
        news += sentence[lbound : m[0]]
        news += hit.format(color, sentence[m[0] : m[1]])
        lbound = m[1]
    news += sentence[matches[-1][1] :]
    return news


def strip_pm(txt, numbers_dict=None):
    txt = txt.lower()
    hoffset = 12 * ('pm' in txt or "afternoon" in txt or 'p.m' in txt)
    txt = txt.replace("afternoon", "")
    txt = txt.replace("'", " ")
    txt = re.sub(".?o.?clock", "", txt)
    if numbers_dict is None:
        parts = re.sub("[:hapm.]+", " ", txt).split()
    else:
        parts = re.sub(" ?[ap][.]?m[.]?$", " ", txt).split()
    parts = [x for x in parts if x]
    microsecond = None
    second = None
    minute = None
    if len(parts) == 4:
        hour, minute, second, microsecond = parts
    elif len(parts) == 3:
        hour, minute, second = parts
    elif len(parts) == 2:
        hour, minute = parts
    else:
        hour = parts[0]
    if numbers_dict is not None:
        hour = numbers_dict[hour]
    if int(hour) == 12 and hoffset:
        hoffset = 0
    hour = (hoffset + int(hour)) % 24
    hour = 0 if hour is None else int(hour)
    minute = 0 if minute is None else int(minute)
    second = 0 if second is None else int(second)
    microsecond = 0 if microsecond is None else int(microsecond)
    return hour, minute, second, microsecond


def erase_level(d, min_level):
    if min_level == Units.YEAR:
        d = d.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.SEASON:
        d = d.replace(day=21, hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.QUARTER:
        d = d.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.MONTH:
        d = d.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.WEEK:
        d = d.replace(hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.DAY:
        d = d.replace(hour=0, minute=0, second=0, microsecond=0)
    elif min_level == Units.HOUR:
        d = d.replace(minute=0, second=0, microsecond=0)
    elif min_level == Units.MINUTE:
        d = d.replace(second=0, microsecond=0)
    elif min_level == Units.SECOND:
        d = d.replace(microsecond=0)
    return d


def num_corrector(start_date, now, words, locale, fallback):
    mod_score = sum([locale.MODIFIERS.get(x, 0) for x in words])
    numeric = [x in locale.ORDINAL_NUMBERS or re.search("^[0.9,.]+$", x) for x in words]
    if mod_score > 0 and numeric:
        end_date = start_date
        start_date = now
    elif mod_score < 0 and numeric:
        end_date = now
    else:
        end_date = start_date + fallback
    return start_date, end_date


def resolve_end_period(start_date, levels, past_boundary, future_boundary, now, words, locale):
    date_changed = False
    min_level = min(levels)
    start_date = erase_level(start_date, min_level)
    if min_level == Units.YEAR:  # year
        end_date = start_date + rd(years=1)
    elif min_level == Units.SEASON:  # season
        end_date = start_date + rd(months=3)
    elif min_level == Units.QUARTER:  # quarter
        for cap in [1, 4, 7, 10]:
            if start_date.month <= cap:
                break
        end_month = cap + 3
        if end_month > 12:
            end_date = start_date + rd(month=end_month % 12, day=1, years=1)
        else:
            end_date = start_date + rd(month=end_month, day=1)
    elif min_level == Units.MONTH:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(months=1))
    elif min_level == Units.WEEK:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(days=7))
    elif min_level == Units.DAY:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(days=1))
    elif min_level == Units.HOUR:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(hours=1))
    elif min_level == Units.MINUTE:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(minutes=1))
    elif min_level == Units.SECOND:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(seconds=1))
    elif min_level == Units.MICROSECOND:
        start_date, end_date = num_corrector(start_date, now, words, locale, rd(microseconds=1))
    # has to be tighter
    # if end_date < past_boundary:
    #     start_date = start_date.replace(year=start_date.year + 1)
    #     end_date = end_date.replace(year=end_date.year + 1)
    #     date_changed = True
    # elif start_date < past_boundary:
    #     start_date = now
    #     date_changed = True
    # note that this breaks far "tomorrow" with future=False
    if start_date > future_boundary:
        for level in ["minutes", "hours", "days", "weeks", "months", "years"]:
            diff = rd(**{level: 1})
            if start_date - diff * 2 < future_boundary:
                start_date = start_date - diff
                end_date = end_date - diff
    elif end_date > future_boundary:
        end_date = future_boundary
    if end_date < past_boundary:
        for level in ["minutes", "hours", "days", "weeks", "months", "years"]:
            diff = rd(**{level: 1})
            if end_date + diff * 2 > past_boundary:
                start_date = start_date + diff
                end_date = end_date + diff
                break
    elif start_date < past_boundary:
        start_date = past_boundary
    return start_date, end_date, date_changed


BOUNDARIES = {1: 31, 2: 29, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}



import itertools
from datetime import datetime

from dateutil.relativedelta import relativedelta as rd
from dateutil.rrule import MONTHLY, rrule

from metadate.utils import erase_level, DAY


"""
# and and or make both go into the interval list
t = "every 2 weeks between this year and the next until December on Tuesdays and Wednesdays at ten and 11 pm for 10 minutes"
mp = MetaPeriod([rd(years=0, weekday=4, hour=10, minute=0), rd(years=0, weekday=5, hour=10, minute=0)],
                rd(years=1, month=12),
                rd(),
                rd(minute=10),
                [rd(weeks=2)])
for s, e in mp.cycle(level=9):
    print(s, e)

r = rrule(WEEKLY, interval=2, dtstart=datetime(2010, 1, 1), byhour=(22, 23), byweekday=(2, 3))
for x in r.between(datetime(2016, 1, 1), datetime(2017, 12, 1)):
    print(x, x + rd(minutes=10))

    # start/end:
#     between ... and ...
#     from ... to ...

# selector:
#     on ...
#     at ...
#     in ...

# highest level - 1
# then has to be constrained to highest_level

# MetaEvery + MetaBetween = every with start and end
# MetaEvery + MetaUntil/For = every with end
# MetaEvery + MetaFrom = every with start and end is final_date


# duration (apply at end to start and end)
#    for ...

t = "Every 3 hours from 9 AM to 5 PM"
mp = MetaPeriod(rd(hour=9),
                rd(hour=17),
                [rd()],
                [rd(hours=1)],
                [rd(hours=3)])
for s, e in mp.cycle(level=3):
    print(s, e)

r = rrule(HOURLY, interval=3, dtstart=datetime(2010, 1, 1))
for x in r.between(datetime(2016, 9, 5, 9), datetime(2016, 9, 5, 17), inc=True):
    print(x, x + rd(minutes=10))


t = "at 10 pm for 10 minutes"
mp = MetaPeriod(rd(hour=10),
                rd(hour=11),
                [rd()],
                [rd(minute=10)],
                None)
for s, e in mp.cycle(level=2):
    print(s, e)

r = rrule(HOURLY, interval=1, dtstart=datetime(2016, 9, 5, 10))
for x in r.between(datetime(2016, 9, 5, 10, 0), datetime(2016, 9, 5, 10, 1), inc=True):
    print(x, x + rd(minutes=10))

t = "on Tuesday at 3 pm for 3 hours"
mp = MetaPeriod(rd(weekday=1),
                rd(weekday=2), r=rrule(HOURLY, interval=1, count=1, dtstart=start, byweekday=(TU), byhour=(15))
                [rd(hour=3) + rd()],
                [rd(hour=3) + rd(hours=3)],
                None)
for s, e in mp.cycle(level=3):
    print(s, e)

start = datetime(2016, 9, 5, 14, 58)
end = datetime(2020, 9, 5, 14, 58)
r = rrule(HOURLY, interval=1, count=1, dtstart=start, byweekday=(TU), byhour=(15))
for x in r.between(start, end, inc=True):
    x = erase_level(x, 3)
    print(x, x + rd(hours=3))


start = datetime(2015, 9, 5, 14, 58)
end = datetime(2020, 9, 5, 14, 58)
r = rrule(MONTHLY, interval=1, dtstart=start, byweekday=(TU), byhour=(15), bysetpos=-1)
for x in r.between(start, end, inc=True):
    x = erase_level(x, 3)
    print(x, x + rd(hours=3))


start = datetime(2015, 9, 5, 14, 58)
end = datetime(2020, 9, 5, 14, 58)
r = rrule(MONTHLY, interval=1, count=1, dtstart=start, bymonthday=(16), byhour=(15), bysetpos=-1)
for x in r.between(start, end, inc=True):
    x = erase_level(x, 3)
    print(x, x + rd(hours=3))


def test_three():
    t = "The 3 of us around $100 (every 2 weeks) at (10:00) (two weeks after 25th of June) in (2018) to Brussels."
    mp = parse_date(t, verbose=VERBOSE)
    assert mp.start_date == dt(2018, 6, 25, 10, 0)
    assert mp.end_date == dt(2018, 6, 25, 10, 1)

from metadate.utils import FREQ


def level_to_duration(level):
    if level == 9:    # year
        return rd(years=1)
    elif level == 8:  # season
        return rd(months=3)
    elif level == 7:  # quarter
        return rd(months=3)
    elif level == 6:  # month
        return rd(months=1)
    elif level == 5:  # week
        return rd(weeks=1)
    elif level == 4:  # day
        return rd(days=1)
    elif level == 3:  # hour
        return rd(hours=1)
    elif level == 2:  # minute
        return rd(minutes=1)
    elif level == 1:  # second
        return rd(seconds=1)
    return rd()


def gen_date(start=None, end=None, freq=None, interval=None, count=None, duration=None, level=2, **kwargs):
    start = start or datetime.now()
    end = end or start + rd(years=3)
    duration = duration or level_to_duration(level)
    if freq is None:
        freq = "yearly"
        interval = 10000
        count = 1
    for x in rrule(freq=FREQ[freq], dtstart=start, interval=interval,
                   until=end, count=count, **kwargs):
        x = erase_level(x, level)
        yield x, x + duration


len(list(gen_date(datetime(2018, 6, 25, 10, 00), freq="weekly", level=2, interval=2)))
len(list(gen_date(datetime(2018, 6, 25, 10, 00), freq="hourly", level=2, interval=2 * 24 * 7)))

list(gen_date(datetime(2018, 6, 25, 10, 00)))
"""


# start = datetime(2017, 6, 1, 0, 0)
# end = datetime(2017, 8, 1, 0, 0)
# r = rrule(DAILY, dtstart=start, byhour=(10))
# for x in r.between(start, end, inc=True):
#     x = erase_level(x, 3)
#     print(x, x + rd(hours=1))

# start = datetime(2017, 1, 28, 0, 0)
# end = datetime(2020, 1, 28, 0, 0)
# r = rrule(MONTHLY, dtstart=start, bymonth=(6), byweekday=(WE), byhour=(11), bysetpos=2)
# for x in r.between(start, end, inc=True):
#     x = erase_level(x, 3)
#     print(x, x + rd(hours=1))


# units = ["quarter", "season", "year", "month", "week", "weekday", "day", "hour", "minute", "second"]

# start_rel =
# duration =
# end_rel =
# interval =
# count =
# freq =

# ["winter 2017", "summer 2018", "2017 days"]


# start = datetime(2017, 1, 28, 0, 0)
# end = datetime(2020, 1, 28, 0, 0)
# r = rrule(MONTHLY, interval=2, dtstart=start, byweekday=(MO), byhour=(10), bysetpos=2)
# for x in r.between(start, end, inc=True):
#     x = erase_level(x, 3)
#     print(x, x + rd(hours=1))


# start = datetime(2017, 1, 28, 0, 0)
# end = datetime(2020, 1, 28, 0, 0)
# r = rrule(MONTHLY, until=end, interval=2, dtstart=start, byweekday=(MO), byhour=(10, 11))
# for x in r:
#     x = erase_level(x, 3)
#     print(x, x + rd(hours=1))

from metadate.utils import FREQ


def level_to_duration(level):
    if level == 9:    # year
        return rd(years=1)
    elif level == 8:  # season
        return rd(months=3)
    elif level == 7:  # quarter
        return rd(months=3)
    elif level == 6:  # month
        return rd(months=1)
    elif level == 5:  # week
        return rd(weeks=1)
    elif level == 4:  # day
        return rd(days=1)
    elif level == 3:  # hour
        return rd(hours=1)
    elif level == 2:  # minute
        return rd(minutes=1)
    elif level == 1:  # second
        return rd(seconds=1)
    return rd()


def gen_date(ref=None, start=None, end=None, freq=None, interval=1, count=None, duration=None,
             level=DAY, **kwargs):
    ref = ref or datetime.now()
    start = start or rd()
    start = ref + start
    end = end or rd(years=3)
    end = start + end
    duration = duration or level_to_duration(level)
    if freq is None:
        freq = "yearly"
        interval = 10000
        count = 1
    for x in rrule(freq=FREQ[freq], dtstart=start, interval=interval,
                   until=end, count=count, **kwargs):
        x = erase_level(x, level)
        yield x, x + duration



import math
from datetime import datetime
from metadate.utils import Units
from metadate.utils import FREQ
from metadate.utils import RD_ARG_TO_UNIT
from metadate.utils import BOUNDARIES
from dateutil.relativedelta import relativedelta


class Meta(object):
    def __init__(self, x, span):
        self.x = x
        self.span = span

    def __repr__(self):
        msg = "{}(x={})"
        return msg.format(self.__class__.__name__, self.x)

    @property
    def is_relative(self):
        return isinstance(self, MetaRelative)


class MetaPeriod(Meta):
    def __init__(
        self, start_date, end_date, levels, spans, locale, text, has_modifier, reference_date
    ):
        self.start_date = start_date
        self.end_date = end_date
        self.levels = levels
        self.spans = spans
        self.locale = locale
        self.text = text
        self.matches = [text[i:j] for i, j in spans]
        self.has_modifier = has_modifier
        self.reference_date = reference_date

    def __repr__(self):
        name = self.__class__.__name__
        cases = ["start_date", "end_date", "levels", "matches", "locale"]
        n = len(name) + 1
        atts = ", ".join(
            [str(getattr(self, x)) for x in cases if getattr(self, x, None) is not None]
        )
        return "{}({})".format(name, atts)

    @property
    def min_level(self):
        return min(self.levels)

    @property
    def max_level(self):
        return max(self.levels)

    @property
    def level(self):
        import warnings

        warnings.simplefilter("once", DeprecationWarning)
        warnings.warn("Please upgrade to using 'levels'", DeprecationWarning)
        return self.min_level

    def __gt__(self, other):
        return self.start_date > other.start_date or (
            self.start_date == other.start_date and self.end_date > self.end_date
        )

    def __eq__(self, other):
        return self.start_date == other.start_date and self.end_date == self.end_date

    def __lt__(self, other):
        return self.start_date < other.start_date or (
            self.start_date == other.start_date and self.end_date < self.end_date
        )

    @property
    def is_in_past(self):
        return self.start_date < self.reference_date

    @property
    def is_publish_date(self):
        return (self.has_day or self.has_modifier) and self.is_in_past

    @property
    def is_today(self):
        return self.start_date.date() == self.reference_date.date()

    @property
    def has_time(self):
        return self.has_second or self.has_minute or self.has_hour

    @property
    def has_second(self):
        return Units.SECOND in self.levels

    @property
    def has_minute(self):
        return Units.MINUTE in self.levels

    @property
    def has_hour(self):
        return Units.HOUR in self.levels

    @property
    def has_day(self):
        return Units.DAY in self.levels

    @property
    def has_month(self):
        return Units.MONTH in self.levels

    @property
    def has_year(self):
        return Units.YEAR in self.levels

    def to_dict(self):
        result = {
            "start_date": self.start_date.isoformat(),
            "end_date": self.end_date.isoformat(),
            "levels": list(self.levels),
            "spans": [{"begin": x[0], "end": x[1]} for x in self.spans],
            "matches": self.matches,
        }
        return result

    @property
    def tz(self):
        return self.start_date.tzinfo


class MetaUnit(Meta):
    def __init__(self, unit, span, modifier=1):
        self.unit = unit
        self.span = span
        self.levels = set([Units[unit]])
        self.modifier = modifier

    def __repr__(self):
        cases = ["unit", "modifier"]
        atts = {x: str(getattr(self, x)) for x in cases if getattr(self, x, None) is not None}
        values = ", ".join(["{}={}".format(x, atts[x]) for x in cases if x in atts])
        return "{}({})".format(self.__class__.__name__, values)


class MetaOrdinal(Meta):
    def __init__(self, amount, span):
        self.amount = float(amount)
        self.span = span

    def __repr__(self):
        cases = ["amount"]
        atts = {x: str(getattr(self, x)) for x in cases if getattr(self, x, None) is not None}
        values = ", ".join(["{}={}".format(x, atts[x]) for x in cases if x in atts])
        return "{}({})".format(self.__class__.__name__, values)


class MetaRelative(Meta):
    def __init__(self, span, levels=None, rd=None, modifier=False, **rd_args):
        if levels is None:
            levels = set(RD_ARG_TO_UNIT[x] for x in rd_args)
        self.rd_args = rd_args
        self.levels = levels
        self.modifier = modifier
        if rd is not None:
            self.rd = rd_args
        else:
            for x in ["seasons", "quarters"]:
                if x in rd_args:
                    del rd_args[x]
            if isinstance(rd_args.get("days"), float) and int(rd_args["days"]) != float(
                rd_args["days"]
            ):
                frac, _ = math.modf(rd_args["days"])
                # currently only halfs are supported
                rd_args["hours"] = rd_args.get("hours", 0) + int(24 * frac)
                rd_args["days"] = int(rd_args["days"])
                self.levels.add(Units.HOUR)
            if isinstance(rd_args.get("months"), float) and int(rd_args["months"]) != float(
                rd_args["months"]
            ):
                frac, _ = math.modf(rd_args["months"])
                # currently only halfs are supported
                rd_args["days"] = rd_args.get("days", 0) + int(30 * frac)
                rd_args["months"] = int(rd_args["months"])
                self.levels.add(Units.DAY)
            if isinstance(rd_args.get("years"), float) and int(rd_args["years"]) != float(
                rd_args["years"]
            ):
                frac, _ = math.modf(rd_args["years"])
                rd_args["days"] = rd_args.get("days", 0) + int(365 * frac)
                rd_args["years"] = int(rd_args["years"])
                self.levels.add(Units.MONTH)
            self.rd = relativedelta(**rd_args)
        self.span = span

    def __add__(self, other):
        # if overlap, crash
        rd = self.rd + other.rd
        levels = self.levels.union(other.levels)
        span = sorted(self.span + other.span)
        modifier = self.modifier or getattr(other, "modifier", False)
        return MetaRelative(span, levels, modifier=modifier, rd=rd)

    def __repr__(self):
        return "{}(rd={}, levels={})".format(self.__class__.__name__, self.rd, self.levels)


class MetaModifier(Meta):
    def __init__(self, x, value, span):
        self.x = x
        self.value = value
        self.span = span

    def __repr__(self):
        msg = "{}(x={}, value={})"
        return msg.format(self.__class__.__name__, self.x, self.value)


class MetaRange(Meta):
    pass


class MetaBetween(Meta):
    def __init__(self, start, end, levels, span):
        self.start = start
        self.end = end
        self.levels = levels
        self.span = span

    def __repr__(self):
        msg = "{}(start={}, end={}, levels={})"
        return msg.format(self.__class__.__name__, self.start, self.end, self.levels)


class MetaEvery(Meta):
    def __init__(self, freq, interval, levels, span):
        self.span = span
        self.freq = freq
        self.interval = interval
        self.levels = levels

    def __repr__(self):
        msg = "{}(freq={}, interval={}, levels={})"
        return msg.format(self.__class__.__name__, self.freq, self.interval, self.levels)


class MetaAnd(Meta):
    pass


class MetaDuration(Meta):
    def __init__(self, rd, levels, span):
        self.span = span
        self.rd = rd
        self.levels = levels

    def __repr__(self):
        msg = "{}(rd={}, levels={})"
        return msg.format(self.__class__.__name__, self.rd, self.levels)



NAME = "en"

# make it not based on index, but key val, and include "hrs", "wks"
UNITS = {
    'microseconde': 'MICROSECOND',
    'microseconden': 'MICROSECOND',
    'second': 'SECOND',
    'seconde': 'SECOND',
    'minuut': 'MINUTE',
    'minuten': 'MINUTE',
    'mins': 'MINUTE',
    'min': 'MINUTE',
    'uur': 'HOUR',
    'uren': 'HOUR',
    'dag': 'DAY',
    'dagen': 'DAY',
    'week': 'WEEK',
    'weken': 'WEEK',
    'kwartaal': 'QUARTER',
    'kwartalen': 'QUARTER',
    'maand': 'MONTH',
    'maanden': 'MONTH',
    'seizoen': 'SEASON',
    'seizoenen': 'SEASON',
    'jaar': 'YEAR',
    'jaren': 'YEAR'
}


MODIFIERS = {
    "in": 1,
    "op": 0,
    "dit": 0,
    "deze": 0,
    "volgend":  1,
    "volgende":  1,
    "aankomend": 1,
    "aankomende": 1,
    "na": 1,
    "voor": -1,
    'recent': -1,
    'afgelope': -1,
    'afgelopen': -1,
    # untested from here
    "vorig": -1,
    "vorige": -1,
    "laatste": -1,
    "laatst": -1,
    "eerste": 1,
    "eerst": 1,
    "geleden": -1,
    "gelede": -1,
    "eerder": -1,
    "eerdere": -1
}

ORDINAL_NUMBERS = {
    'nul': 0,
    'een': 1,
    'twee': 2,
    'drie': 3,
    'een paar': 3,
    'weinig': 3,
    'vier': 4,
    'vijf': 5,
    'zes': 6,
    'zeven': 7,
    'acht': 8,
    'negen': 9,
    'tien': 10,
    'elf': 11,
    'twaalf': 12,
    'dertien': 13,
    'veertien': 14,
    'vijftien': 15,
    'zestien': 16,
    'zeventien': 17,
    'achttien': 18,
    'negentien': 19,
    'twintig': 20,
    'dertig': 30,
    'veertig': 40,
    'vijftig': 50,
    'zestig': 60,
    'zeventig': 70,
    'tachtig': 80,
    'negentig': 90,
    'honderd': 100,
    'duizend': 1000,
}


IN_THE = ['in de', 'over de', 'over het', 'tijdens de', 'tijdens het', 'binnen de', 'binnen het']


MONTHS = {
    "januari": 1, "februari": 2, "maart": 3, "april": 4, "mei": 5, "juni": 6, "juli": 7,
    "augustus": 8, "september": 9, "oktober": 10, "november": 11, "december": 12
}
MONTHS_SHORTS = {k[:3]: v for k, v in MONTHS.items()}
MONTHS_SHORTS["febr"] = 2
MONTHS_SHORTS["sept"] = 9

WEEKDAY = {
    "maandag": 0,
    "dinsdag": 1,
    "woensdag": 2,
    "donderdag": 3,
    "vrijdag": 4,
    "zaterdag": 5,
    "zondag": 6
}

WEEKDAY_SHORTS = {k[:3]: v for k, v in WEEKDAY.items()}

SEASONS = {
    "winter": 12,
    "lente": 3,
    "zomer": 6,
    "herfst": 9
}

QUARTERS = {
    "q1": 1,
    "eerste kwartaal": 1,
    "q2": 4,
    "tweede kwartaal": 4,
    "q3": 7,
    "derde kwartaal": 7,
    "q4": 10,
    "vierde kwartaal": 10,
    "laatste kwartaal": 10
}

NOW = [
    'nu'
]

TODAY_TOMORROW = {
    "vandaag": 0,
    "morgen": 1,
    "gisteren": -1,
    "overmorge": 2,
    "overmorgen": 2,
    "eergister": -1,
    "eergisteren": -1
}

NOON = {
    "middag": 12,
    "middernacht": 0,
    "eind van de middag": 17,
    "eind van de ochtend": 11,
}

AND = ["en", "of", "tot"]

RANK_NAMING = ['e']
ON_THE = ['op de']
AT = ['om']

WHITELIST = ["op", "of", "van", ",", "de", "nacht",
             "morgen", "voor", "om", "middag", "ochtend", "avond"]


BETWEEN = ["tussen", "voor"]
EVERY = ["elke"]
DURATION = ["voor", "tot"]

DD_LEFT_FIRST = True



from setuptools import find_packages
from setuptools import setup

MAJOR_VERSION = "0"
MINOR_VERSION = "5"
MICRO_VERSION = "76"
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(
    name='metadate',
    version=VERSION,
    description="All your date are belong to us",
    author='Pascal van Kooten',
    url='https://github.com/kootenpv/metadate',
    author_email='kootenpv@gmail.com',
    install_requires=['python-dateutil', 'enum-compat'],
    entry_points={'console_scripts': ['metadate = metadate.__main__:main']},
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Customer Service',
        'Intended Audience :: System Administrators',
        'Operating System :: Microsoft',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    platforms='any',
)



import re
from datetime import datetime

from dateutil.relativedelta import relativedelta as rd

from metadate.classes import (
    MetaBetween,
    MetaModifier,
    MetaOrdinal,
    MetaPeriod,
    MetaRange,
    MetaRelative,
    MetaUnit,
)
from metadate.scanner import Scanner
from metadate.utils import Units, BOUNDARIES, erase_level, log, resolve_end_period
from metadate.classes import MetaAnd

import metadate.locales.en as locale_en
import metadate.locales.nl as locale_nl
from metadate.add_timezone import add_timezone

# TODO: have to make locale lazy still

lang_to_locale = {"en": locale_en, "nl": locale_nl}
lang_to_scanner = {}


def get_relevant_parts(matches):
    strike = 0
    bundles = [[]]
    for m in matches:
        if isinstance(m, str):
            strike = 1
        else:
            if strike:
                bundles.append([m])
            else:
                bundles[-1].append(m)
            strike = 0
    return bundles


def between_allowed(x, y, text):
    start = x.span[1]
    end = y.span[0]
    return re.match("^[ -]*(and)?[ -]*$", text[start:end])


def merge_ordinal_unit(matches, text):
    news = []
    t = 0
    last = False
    n = len(matches)
    spans = []
    for i, m in enumerate(matches):
        if i != n - 1 and isinstance(m, MetaOrdinal):
            if last and not between_allowed(last, m, text):
                t = 0
            t += float(m.amount)
            last = m
            spans.append(m.span)
            continue
        elif isinstance(m, MetaUnit):
            m.modifier *= t if t else 1
            m.span = [m.span] + spans
            spans = []
        news.append(m)
    return news


def cleanup_relevant_parts(bundles, locale, now):
    cleaned_bundles = []
    for bundle in bundles:
        modifier = False
        relative = False
        meta_units = []
        new = []
        for x in bundle:
            if isinstance(x, MetaModifier):  # "this"
                if meta_units:
                    for mu in meta_units:
                        unit = mu.unit.lower() + 's'
                        rd_kwargs = {unit: mu.modifier * locale.MODIFIERS[x.x]}
                        new.append(
                            MetaRelative(
                                levels=set([Units[mu.unit]]),
                                span=[x.span, mu.span],
                                modifier=True,
                                **rd_kwargs,
                            )
                        )
                if relative and relative.rd.weekday is not None:
                    relative.modifier = x
                modifier = x
                relative = False
                meta_units = []
            elif isinstance(x, MetaAnd):
                continue
            elif modifier:
                if isinstance(x, MetaUnit):
                    if x.unit == "quarter":
                        rd_kwargs = {
                            "months": 3 * x.modifier * locale.MODIFIERS[modifier.x],
                            "day": 1,
                        }
                    # elif x.unit == "season":
                    #     rd_kwargs = {"months": 3 * x.modifier * locale.MODIFIERS[modifier.x],
                    #                  "day": 21}
                    else:
                        unit = x.unit.lower() + 's'
                        rd_kwargs = {unit: x.modifier * locale.MODIFIERS[modifier.x]}
                    new.append(
                        MetaRelative(
                            levels=set([Units[x.unit]]),
                            span=[x.span, modifier.span],
                            modifier=True,
                            **rd_kwargs,
                        )
                    )
                # elif isinstance(x, MetaDate):
                #     # if hasattr(x, "month"):
                #     #     print("relative", x.month, modifier)
                #     # if hasattr(x, "season"):
                #     #     print("relative", x.season, modifier)
                #     new.append(x)
                elif x.is_relative:
                    if x.rd.weekday is not None:
                        modifier_value = locale.MODIFIERS[modifier.x]
                        merge_rd_modifier(x, modifier_value, now)
                        modifier = False
                        relative = False
                    new.append(x)
            elif isinstance(x, MetaUnit):
                meta_units.append(x)
                modifier = False
                relative = False
            elif isinstance(x, MetaOrdinal):
                continue
            elif isinstance(x, MetaRelative):
                relative = x
                new.append(x)
            else:
                new.append(x)
        cleaned_bundles.append(new)
    cleaned_bundles = [x for x in cleaned_bundles if any([y.is_relative for y in x])]
    return cleaned_bundles


def flatten_inner(l):
    span = []
    for s in l:
        if isinstance(s, list):
            span.extend(flatten_span(s))
        else:
            span.append(s)
    return span


def flatten_span(l):
    return sorted(set(flatten_inner(l)))


def get_levels(cleaned_bundle):
    levels = set()
    for x in cleaned_bundle:
        if isinstance(x, (MetaRelative, MetaUnit)):
            levels.update(x.levels)
    return min(levels), levels


def correct_rd_weekday(cleaned_bundle, now):
    for x, y in zip(cleaned_bundle[:-1], cleaned_bundle[1:]):
        if x.is_relative and x.rd.weekday is not None and x.rd.days == 0:
            relative_weekday = rd(weekday=x.rd.weekday)
            if y.is_relative:
                now = now + y.rd
            now_weekday = now.weekday()

            # avoids 0/1 index problem with day
            same_day = (now + relative_weekday).weekday() == now_weekday
            still_to_come_this_week = (now + relative_weekday).weekday() > now_weekday
            already_passed_this_week = (now + relative_weekday).weekday() < now_weekday

            if same_day and x.modifier:
                if x.modifier.value > 0:
                    x.rd.weeks = 1
                else:
                    x.rd.weeks = -1
            # on wednesday saying: "monday in 2 weeks"
            if x.modifier and already_passed_this_week:
                x.rd.weeks = -1


def merge_rd_modifier(x, modifier_value, now):
    # x.modifier = True
    relative_weekday = rd(weekday=x.rd.weekday)
    now_weekday = now.weekday()

    # avoids 0/1 index problem with day
    same_day = (now + relative_weekday).weekday() == now_weekday
    still_to_come_this_week = (now + relative_weekday).weekday() > now_weekday
    already_passed_this_week = (now + relative_weekday).weekday() < now_weekday
    # relativedelta weekdays added always choose the next one
    # so when the modifier value is 1 week then ignore it
    if still_to_come_this_week and modifier_value == 1:
        modifier_value = 0
    if already_passed_this_week and modifier_value == 1:
        modifier_value = 0
    if same_day and modifier_value == 0:
        modifier_value = 1
    x.rd.weeks += modifier_value


def datify(cleaned_bundle, past_boundary, future_boundary, now, locale, text):
    span = flatten_span([x.span for x in cleaned_bundle])
    min_level, levels = get_levels(cleaned_bundle)
    words = [text[x[0] : x[1]].lower() for x in span]
    # print(level)
    rdt, erdt = resolve_dt(cleaned_bundle, now)

    correct_rd_weekday(cleaned_bundle, now)

    if erdt is not None:
        start_date = rdt + resolve_rd(cleaned_bundle)
        end_date = erdt + resolve_rd(cleaned_bundle)
        start_date, _, _ = resolve_end_period(
            start_date, levels, past_boundary, future_boundary, now, words, locale
        )
        end_date = erase_level(erdt, min_level)
    else:
        resolved_rd = resolve_rd(cleaned_bundle)
        try:
            start_date = rdt + resolved_rd
        except TypeError:
            resolved_rd.years = int(resolved_rd.years)
            resolved_rd.months = int(resolved_rd.months)
            try:
                start_date = rdt + resolved_rd
            except ValueError:
                return None, None, None, None
        except ValueError:
            return None, None, None, None
        start_date, end_date, _ = resolve_end_period(
            start_date, levels, past_boundary, future_boundary, now, words, locale
        )
    return start_date, end_date, levels, span


def has_between(bundle):
    return any(isinstance(x, MetaBetween) for x in bundle)


def merge_dt(mdt, dt):
    # mdt is a made-up mutable dt
    # month is precomputed so that day can be init
    month = mdt[1] if mdt[1] > -1 else dt.month
    day = mdt[2] if mdt[2] > -1 else dt.day
    ndt = datetime(
        year=mdt[0] if mdt[0] > -1 else dt.year,
        month=month,
        day=min(day, BOUNDARIES[month]),
        hour=mdt[3] if mdt[3] > -1 else dt.hour,
        minute=mdt[4] if mdt[4] > -1 else dt.minute,
        second=mdt[5] if mdt[5] > -1 else dt.second,
        microsecond=mdt[6] if mdt[6] > -1 else dt.microsecond,
    )
    return ndt


def resolve_dt(cleaned_bundle, now):
    # the concept of checking whether there is overlap is good, and crash?
    indices = {
        'year': 0,
        'month': 1,
        'day': 2,
        'hour': 3,
        'minute': 4,
        'second': 5,
        'microsecond': 6,
    }
    dts = [-1, -1, -1, -1, -1, -1, -1]
    edts = [-1, -1, -1, -1, -1, -1, -1]
    contains_between = has_between(cleaned_bundle)
    # for d in cleaned_bundle:
    #     if isinstance(d, MetaDate):
    #         for x in d.__dict__:
    #             if x not in ['level', 'span', 'now']:
    #                 if dts[indices[x]] != -1:
    #                     if not contains_between:
    #                         raise ValueError("Field already known in dt, should not overwrite?")
    #                 else:
    #                     dts[indices[x]] = getattr(d, x)
    #                 edts[indices[x]] = getattr(d, x)
    dt = merge_dt(dts, now)
    edt = merge_dt(edts, now) if contains_between else None
    return dt, edt


def resolve_rd(cleaned_bundle):
    rds = rd()
    for d in cleaned_bundle:
        if d.is_relative:
            other = d.rd
            # "on sunday 12th of may" --- adding weekday + day should
            # remove corrective weekday +7 days
            # if rds.weekday is not None and other.day is not None:
            #     rds.days = 0
            # if other.weekday is not None and rds.day is not None:
            #     other.days = 0
            rds += other
    return rds


def has_modifier(cleaned_bundle):
    return any([x.modifier for x in cleaned_bundle if x.is_relative])


def handle_meta_range(cleaned_bundle, past_boundary, locale, text, future_boundary, now):
    phase = 0
    mrange = None
    relatives = []
    metadates = []
    contains_modifier = False
    for x in cleaned_bundle:
        if phase == 0 and isinstance(x, MetaRange):
            phase = 1
            mrange = x
        elif phase == 1 or phase == 2 and x.is_relative:
            phase = 2
            relatives.append(x)
            contains_modifier = contains_modifier or x.modifier
        elif phase == 2 or phase == 3 and isinstance(x, MetaPeriod):
            phase = 3
            metadates.append(x)
    if phase < 2:
        return None
    # case 1: MetaRange, MetaRelative, MetaRelative, etc
    # log("relatives", relatives, True)
    # log("metadates", metadates, True)
    if not metadates:
        # in this case, the start_date becomes "now" adjusted for level
        # the end_date is the start_date + relativedeltas following
        min_level, levels = get_levels(relatives)
        rds = rd()
        start_date = erase_level(now, 1)
        for x in relatives:
            if x.rd is None:
                continue
            rds += x.rd
        try:
            end_date = start_date + rds
        except TypeError:
            rds.years = int(rds.years)
            rds.months = int(rds.months)
            end_date = start_date + rds
        span = flatten_span([mrange.span, [x.span for x in relatives]])
        if start_date > end_date:
            start_date, end_date = end_date, start_date
        return MetaPeriod(
            start_date, end_date, levels, span, locale.NAME, text, contains_modifier, now
        )
    else:
        spans = [flatten_span(x.span) for x in relatives]
        words = [text[x[0] : x[1]].lower() for x in spans]
        # generic
        dt, _ = resolve_dt(metadates, now)
        min_dt_level, dt_levels = get_levels(metadates)
        min_rd_level, rd_levels = get_levels(relatives)
        if "first" in words:
            # this is actually the logic for !first! days of, not "next days of"
            # case 2: MetaRange, MetaRelative, MetaRelative, etc, MetaDate, Metadate, etc
            start_date, _, date_changed = resolve_end_period(
                dt, dt_levels, past_boundary, future_boundary, now, words, locale
            )
            end_date = erase_level(dt, min_dt_level) + resolve_rd(relatives)
        elif "last" in words:
            _, end_date, date_changed = resolve_end_period(
                dt, dt_levels, past_boundary, future_boundary, now, words, locale
            )
            start_date = erase_level(end_date, min_dt_level) + resolve_rd(relatives)
        else:
            return None
            raise NotImplementedError("What's the case here?")
    if date_changed:
        start_date = start_date + rd(years=1)  # I think also here?
        end_date = end_date + rd(years=1)
    span = flatten_span([mrange.span] + [x.span for x in relatives] + [x.span for x in metadates])
    return MetaPeriod(
        start_date,
        end_date,
        rd_levels.union(dt_levels),
        span,
        locale.NAME,
        text,
        contains_modifier,
        now,
    )


def min_span(obj):
    span = getattr(obj, "span")
    if span is None:
        return 2 ** 50
    if isinstance(span, list):
        return span[0][0]
    return span[0]


def parse_date(
    text,
    future=rd(years=30),
    past=rd(years=100),
    lang="en",
    reference_date=None,
    scanner=None,
    multi=False,
    verbose=False,
    min_level=None,
    max_level=None,
    return_early=0,
):
    now = reference_date or datetime.now()
    past = past or rd()
    future = future or rd()
    locale = lang_to_locale[lang]
    log("\nSentence", text, verbose)
    if lang not in lang_to_scanner:
        scanner = Scanner(locale)
        lang_to_scanner[lang] = scanner
    scanner = lang_to_scanner[lang]
    matches, _ = scanner.scan(text)
    log("matches", matches, verbose=verbose)
    parts = get_relevant_parts(matches)
    log("1", parts, verbose=verbose)
    if return_early == 1:
        return parts
    merged = [merge_ordinal_unit(x, text) for x in parts]
    if return_early == 2:
        return merged
    log("2", merged, verbose=verbose)
    cleaned_parts = cleanup_relevant_parts(merged, locale, now)
    if return_early == 3:
        return cleaned_parts
    log("3", cleaned_parts, verbose)

    if not cleaned_parts:
        return [] if multi else None

    cleaned_parts = sorted(cleaned_parts, key=len, reverse=True)

    mps = []
    past_boundary = now - past
    future_boundary = now + future

    for cleaned_bundle in cleaned_parts:
        if mps and mps[0] and not multi:
            return add_timezone(mps[0])
        handle_meta_result = handle_meta_range(
            cleaned_bundle, past_boundary, locale, text, future_boundary, now
        )
        if handle_meta_result:
            log("4meta", handle_meta_result, verbose)
            mps.append(handle_meta_result)
            continue

        start_date, end_date, levels, span = datify(
            cleaned_bundle, past_boundary, future_boundary, now, locale, text
        )

        if start_date is None:
            continue

        contains_modifier = has_modifier(cleaned_bundle)
        mp = MetaPeriod(
            start_date, end_date, levels, span, locale.NAME, text, contains_modifier, now
        )

        if min_level is not None and mp.min_level < min_level:
            continue

        if max_level is not None and mp.max_level > max_level:
            continue

        log("4default", mp, verbose)
        if past_boundary <= mp.start_date <= future_boundary:
            mps.append(mp)

    if not multi:
        if mps:
            return add_timezone(mps[0])
        else:
            return None

    return [add_timezone(x) for x in mps]
